/* KernelDensityEstimator.java
 * This source file is part of the DDPref software suite that implements the
 * DDPref language for preferences over sets of objects, and algorithms
 * for learning such preferences automatically from training data.
 * 
 * Copyright (C) 2006  Eric Eaton, Marie desJardins, Kiri Wagstaff
 * 
 * If you use or extend this software, please cite it as:
 *   Marie desJardins, Eric Eaton, and Kiri Wagstaff.  (2006).
 *   Learning User Preferences for Sets of Objects.
 *   In Proceedings of the 23rd International Conference
 *   on Machine Learning, June 25--29, Pittsburgh, PA.
 *  
 * This work was partly supported by NSF grant ITR #0325329 and was partly carried
 * out at the Jet Propulsion Laboratory, California Institute of Technology,
 * under a contract with the National Aeronautics and Space Administration.
 * 
 * Developed at the University of Maryland Baltimore County,
 *                  Dept. of Computer Science and Electrical Engineering,
 *                  MAPLE Laboratory (http://maple.cs.umbc.edu).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

/*
 * @author Eric Eaton
 */

package ddpref;

import java.text.DecimalFormat;
import java.util.*;

import org.apache.commons.math.stat.StatUtils;

import weka.core.*;
import weka.estimators.Estimator;

/** 
 * Kernel density estimator. 
 * Based on Representing Data Distributions with Kernel 
 * Density Estimates.  AMC Tech. Brief. Analytical Methods 
 * Committee.  No. 4. 2001.  Royal Society of Chemistry.
 *
 * @author Eric Eaton
 * @version $Revision: 1.11 $
 */
public class KernelDensityEstimator implements Estimator {

	/** Vector containing all of the values seen */
	protected ArrayList<Double> m_values;
	
	/** Distribution specification */
	protected double[] m_distributionProbs = null;
	protected double m_minXCoord = Double.NaN;
	protected double m_maxXCoord = Double.NaN;
	protected double m_maxProbability = 0;
	protected double m_peak = Double.NaN;
	protected double m_offset = Double.NaN;
	protected double m_halfOffset = Double.NaN;
	protected double m_bandwidth = Double.NaN;
	
	protected HashMap<Double,Double> cache = new HashMap<Double,Double>();


	/**
	 * Constructor
	 */
	public KernelDensityEstimator(double precision) {
		m_values = new ArrayList<Double>();
		m_offset = precision;
		m_halfOffset = precision / 2.0;
	}

	/**
	 * Add a new data value to the current estimator.
	 *
	 * @param data the new data value 
	 * @param weight the weight assigned to the data value (ignored)
	 */
	public void addValue(double data, double weight) {
		double roundedData = Math.rint(data / m_offset) * m_offset;  // round v to the precision
		m_values.add(roundedData);
		m_distributionProbs = null; // reset the distribution
		cache.clear();
		
	}
	
	/** Gets the maximum probability value assigned by the distribution
	 * @return the max probability of a value in this distribution
	 */
	public double getMaxProbability() {
		// generate the distribution if it hasn't been generated
		if (m_distributionProbs == null) {
			generateDistribution(); 
		}
		// return the max probability
		return m_maxProbability;
	}
	
	public double getPeak() {
		return m_peak;
	}
	
	/** Generates the distribution from the given values and weights
	 *
	 */
	private void generateDistribution() {
		
		// if the distribution has already been generated, return
		if (m_distributionProbs != null) {
			return;
		}
		
		// convert the values to a sorted array
		double values[] = new double[m_values.size()];
		int i=0;
		for (double d : m_values) { values[i++] = d; }
		java.util.Arrays.sort(values);
		
		if (values.length == 1) {
			System.out.print("");
		}
		
		// compute the statistics for the data set
		double standardDev = Math.sqrt(StatUtils.variance(values));
		double irq = StatUtils.percentile(values,75) - StatUtils.percentile(values,25);
		standardDev = Math.max(standardDev,m_offset);  // make sure the stdev isn't 0 (for all values the same)
		irq = Math.max(irq,m_offset);  // make sure the stdev isn't 0 (for all values the same)
		
		// compute the bandwidth
		m_bandwidth = 0.9 * Math.min(standardDev, irq / 1.34) 
			* Math.pow((double)m_values.size(), -0.2);
		
		// compute the min and max coordinates
		m_minXCoord = Math.floor(values[0] - 3.0*standardDev);
		m_maxXCoord = Math.ceil(values[values.length-1] + 3.0*standardDev);
		
		// prepare to track the max probability
		m_maxProbability = Double.NEGATIVE_INFINITY;
		
		// generate the probabilities over this range
		int numBins = (int) Math.ceil((m_maxXCoord - m_minXCoord) / m_offset) + 1;
		m_distributionProbs = new double[numBins];
		for (i = 0; i < numBins; i++) {
			double x = m_minXCoord + i*m_offset;
			m_distributionProbs[i] = computeProbability(x);
			if (m_distributionProbs[i] > m_maxProbability) {
				m_maxProbability = m_distributionProbs[i];
				m_peak = x;
			}
		}
		
		if (m_maxProbability == 0) {
			System.err.println("Error in KernelDenstiyEstimator.generateDistribution():  m_maxProbability is 0.");
			System.err.print("   values = [");
			for (double d : values) System.err.print(d+", ");
			System.err.println("]");
			Thread.dumpStack();
		}
		
	}
	
	/** Computes the probability of the specified value
	 * 
	 * @param value The value for the x-coordinate.
	 * @return The probability of that value.
	 */
	private double computeProbability(double value) {
		
		// compute the probability for this value
		double prob = 0;
		for (double v : m_values) {
			double val = (value - v) / m_bandwidth;
			prob += Statistics.normalProbability(val + m_halfOffset) 
				- Statistics.normalProbability(val - m_halfOffset);
		}
		
		// normalize the probability
		prob /= ((double)m_values.size()) * m_bandwidth;
		
		return prob;
		
	}

	/**
	 * Get a probability estimate for a value.
	 *
	 * @param data the value to estimate the probability of
	 * @return the estimated probability of the supplied value
	 */
	public double getProbability(double data) {
		
		//Try to get it from the cache first
		Double dataObj = new Double( data );
		Double cachedVal = cache.get(dataObj);
		if( cachedVal != null )
			return cachedVal.doubleValue();
		
		// generate the distribution if it isn't already
		generateDistribution();
		
		// return zero if it is outside the distribution
		if (data < m_minXCoord || data > (m_maxXCoord-m_offset)) {
			return 0;
		}
		
		// figure out the index into the distribution
		double roundedValue = Math.rint(data / m_offset) * m_offset;
		int idx = (int) Math.rint((roundedValue - m_minXCoord) / m_offset);
		idx = (roundedValue <= data) ? idx : idx-1;  // we want the idx'ed value to be less than the data
		
		// interpolate the probability
		// y =  slope * delta_x + x_value
		// y = (([idx+1]-[idx])/offset)*(x-(min+offset*idx))+[idx]
		double interpolatedProb = (m_distributionProbs[idx+1] - m_distributionProbs[idx])/m_offset *
			(data - (m_minXCoord+m_offset*idx)) + m_distributionProbs[idx];
		
		//cache the result for future use
		cache.put(dataObj, new Double(interpolatedProb) );
		
		return interpolatedProb;
	}
		
	

	/** Display a representation of this estimator */
	public String toString() {
		
		// generate the distribution if it isn't already
		generateDistribution();
		
		String s = "";
		DecimalFormat df = new DecimalFormat("0.0000");
		for (double i=m_minXCoord; i<=m_maxXCoord; i += m_offset) {
			s = s + "("+df.format(i)+","+df.format(getProbability(i))+") ";
		}
		return s;
	}

	/**
	 * Main method for testing this class.
	 *
	 * @param argv should contain a sequence of numeric values
	 */
	public static void main(String[] argv) {

		try {
			if (argv.length < 2) {
				System.out.println("Please specify a set of instances.");
				return;
			}
			KernelDensityEstimator newEst = new KernelDensityEstimator(0.1);
			for (int i = 0; i < argv.length - 3; i += 2) {
				newEst.addValue(Double.valueOf(argv[i]).doubleValue(), Double
						.valueOf(argv[i + 1]).doubleValue());
			}
			System.out.println(newEst);

			double start = Double.valueOf(argv[argv.length - 2]).doubleValue();
			double finish = Double.valueOf(argv[argv.length - 1]).doubleValue();
			for (double current = start; current < finish; current += (finish - start) / 50) {
				System.out.println("Data: " + current + " "
						+ newEst.getProbability(current));
			}
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
	}
}
